
#ifndef __MY_TRAVERSE__
#define __MY_TRAVERSE__

#include "rose.h"
#include "dp.h"
#include "util.h"
#include "dataStruct.h"

using namespace std;
 
#define IN_PROC(s, v, id) s << v << "@" << id

class ModelPF {
 public:

  int num_of_threads;
  int thread_id;        // current thread id
  int barrier_count;    // number of barriers

  bool check_race;      // need to check races?
  bool check_deadlock;  // need to check deadlock?

  DP* dp;               // the decision procedure

  typedef list<Expr> ExprList;
  typedef set<Expr> ExprSet;
  typedef map<string, ExprList*> VarListMap;

 private:

  bool DEBUG_MODE;

  Expr default_trans;  // the default transition when no valid transition is present
  Expr NU_num;         // a never used naturual number, for race check

  typedef map<string, int> VarNumMap;
  typedef map<string, Variable> VarLookupMap;

  VarLookupMap lookupVars;          // look at a variable according to its name (a string)
  VarNumMap sharedVars;             // map a global variable to the miminal unused number

  void init() {
    thread_id = 0;
    // schedule_len = 2;
    
    barrier_count = 0;
    check_race = true;
    check_deadlock = true;
    
   
    default_trans = dp->mk_true();
    NU_num = dp->mk_int_num(100); // during encoding, we didn't assign 100 again.  
  }

  public:

  ModelPF(DP* dp_in, bool debug = true){
    DEBUG_MODE = debug;
    dp = dp_in;
    init();
  };

  ~ModelPF() {     // release memory
    for (VarListMap::iterator ii = sharedTerms.begin(); ii != sharedTerms.end(); ii++) {
      delete (*ii).second; 
    }
  };

  Expr sum_constraints(ExprList& constraints);  // the conjunction of all constraints in the list
  Expr sum_constraints(stack<Expr> constraints);  // the conjunction of all constraints in the stack
  void print_constraints(ExprList& constraints);  // for debugging

  VarNumMap localVars;              // map a local variable to the miminal unused number

  inline bool isSharedVar(string v) {
    return lookupVars[v].is_shared;
  };

  void record_variable(SgNode* refExp); // record the variables
  void add_constraint(Expr e);          // add into the constraint list

  ExprList constraints;      // non path sensistive constraints
  ExprList ps_constraints;   // path sensitive constraints
  // only shared variables are considered for races
  VarListMap sharedTerms;    // shared variables and barriers
 
  ExprList bufferMessage; // the message  buffer stored in constraints 
  ExprList properties;       // the constraitns for the properties to be checked
  map<string, ExprList> barriers; // global barriers, currently both of omp barrier and MPI_Barriers are considerred, but they are not in same node class, SgPragmaStatement and SgFunctionCall

  typedef map<string, list<set<Expr> > > VarOrderMap;
  VarOrderMap sharedVarOrder;   // order the scedule ids of shared variables

  // generate a DP expression for an expression rooted at sgNode
  Expr convertExp(SgNode* sgNode, map<string,Expr>* vmap = NULL);
  bool isBuiltInExp(SgNode* sgNode, Expr &t);

  Expr makeGlobalVarIntactTrans(string v);  // v[schedule_id] = v[schedule_id - 1]

  Expr make_update_array_trans(SgNode* refExp, Expr value);
  // generate the transition for a statement
  Expr buildTrans(SgNode* sgNode, SgNode* refExp, Expr& lhs);

  void set_deadlock_check(string v, Expr schedule_id)
 {
   return; 
   //edited by Hongyi 
   // thoughts are similar like race check, only compare with the schedule ID on  lock_var (v), if there are both schedule for lock_var and unlock_var, then no deadlock, otherwise, report deadlock for this lock resource. 
  
 }
  // record the id of the thread accessing this global variable
 
  void set_access(string v, Expr schedule_id)
  {
    return;

    // could be a read or a write
    Expr rw = dp->get_var_from_name("$" + v);
    Expr rw_var = dp->mk_unary_app(rw, schedule_id);
    Expr rw_id_exp = dp->mk_eq(rw_var, dp->mk_int_num(thread_id)); // map thread t id to var with schedule id
    properties.push_back(rw_id_exp);
  }
  // only interested in the schedule trace
  // record the id of the thread accessing this global variable
  void set_race_check(string v, Expr schedule_id)
  {
    return;

    Expr rw = dp->get_var_from_name("$" + v);
    Expr previous_id = dp->mk_sub2(schedule_id, dp->mk_int_num(1), true);
    Expr old_rw_var = dp->mk_unary_app(rw, previous_id);
    // access from another thread indicate a race in a same varible
    //example : t1 ->W(v) at (10), t2->W(v) at (11)
    Expr race_exp1 = dp->mk_diseq(old_rw_var, dp->mk_int_num(thread_id));
    Expr race_exp2 = dp->mk_diseq(old_rw_var, NU_num);
    Expr race_exp = dp->mk_and2(race_exp1, race_exp2);
    properties.push_back(race_exp);
  }

  Expr mk_local_var(string name, VarNumMap& varMap);
  Expr mk_unchanged_trans(Expr var, Expr schedule_id);

  // customize the construction of a (global) variable
  Expr construct_var(SgInitializedName* initName, map<string,Expr>* vmap) {  // construct a term v[index]
    string s = initName->get_name().str();
    Variable v = lookupVars[s];
    if (v.is_shared) {
      Expr schedule_id;
      map<string,Expr>::iterator ii = vmap->find(s);
      if (ii == vmap->end()) {
	ROSE_ASSERT(sharedTerms[s]->size() > 0);
	schedule_id = sharedTerms[s]->back();
      }
      else {
	schedule_id = ii->second;
      }
      // build schedule id with name and ssa
      return dp->mk_unary_app(v.get_var_expr(dp), schedule_id);
    }
    else
      return fetch_update_var(v);
  }

  Expr fetch_update_var(Variable& v, bool redefine_local = false, bool redefine_shared = false, 
			bool for_shared_lhs = false) {
    string name = v.name;
    stringstream out;
	
    if (v.is_shared)   // is a shared variable?
      {
	cout << "shared_var = " << v << endl;

	if (redefine_shared || for_shared_lhs) 
	    sharedVars[name]++;

	out << IN_PROC(name, sharedVars[name], thread_id);  // schedule id 
	name = out.str();
	
	Expr schedule_id = dp->mk_var(name, dp->mk_nat_type());
	Expr var_name = v.get_var_expr(dp);
	Expr new_var = dp->mk_unary_app(var_name, schedule_id);

	if (check_race) {
	  set_access(name, schedule_id);
	}

	if (!redefine_shared && !for_shared_lhs) {  // 
	  return new_var;
	}

	sharedTerms[v.name]->push_back(schedule_id);
	set<Expr> id_set;
	id_set.insert(schedule_id);
	sharedVarOrder[v.name].push_back(id_set);     // the order for shared variables

	if (redefine_shared)
	{
	  // if (sharedVars[v.name] != 1) 
	  {
	    Expr update_trans = mk_unchanged_trans(var_name, schedule_id);

	    dp->pp_expr(update_trans);
	    cout << endl;

	    add_constraint(update_trans);  // don't guard this transition with conditions
	  }
	  return new_var;
	}

	if (for_shared_lhs) {  // a write
	  // check the race
	  if (check_race) 
	    set_race_check(name, schedule_id);
	  return new_var;
	}
      }
    else {       // a local variable
      out << IN_PROC (name, localVars[name], thread_id);
      name = out.str();
      Expr var = dp->mk_var(name, v.get_var_type(dp));
      return var;
    }
  };



  // then verify race inside
  // return the SSA format of a variable
  Expr obtain_SSA_var(SgInitializedName* initName, bool redefine_shared = true) {
    // SgInitializedName* initName = varRefExp->get_symbol()->get_declaration();
    string s = initName->get_name().str();
    return fetch_update_var(lookupVars[s], false, redefine_shared);
  }

  // update a variable's SSA information
  Expr redefine_var(SgInitializedName* initName) {
    // SgInitializedName* initName = varRefExp->get_symbol()->get_declaration();
    string s = initName->get_name().str();
    Variable v = lookupVars[s];
    if (!v.is_shared) { 
	localVars[s]++;
    }
      // the variable in the left hand side should be redefined
    return fetch_update_var(v, true, false, true); 
  }
  
};

string formatOP (SgNode* sgNode);
string formatExp(SgNode* sgNode);


#endif
